from funkyprompt.core.agents import CallingContext
from funkyprompt.core.agents import MessageStack
import typing


class LanguageModelBase:

    def __call__(
        self,
        messages: str | typing.List[dict] | MessageStack,
        context: CallingContext = None,
        functions: typing.Optional[dict] = None,
        **kwargs
    ):
        """the callable is a lightly more opinionated version of run for convenience
        but users of run should remain close to what the model needs
        """
        from funkyprompt.core.agents import DefaultAgentCore

        context = context or CallingContext()
        if isinstance(messages, str):
            """simple convenience cheat"""
            messages = MessageStack(question=messages, model=DefaultAgentCore)
        if isinstance(messages, MessageStack):
            messages = messages.model_dump()

        self._messages = messages
        self._functions = functions

        return self.run(messages=messages, context=context, functions=functions)


from .gpt import GptModel


def language_model_client_from_context(
    context: CallingContext = None, with_retries: int = 0
):
    """The model is loaded from the context.
    Retries are used sparingly for some system contexts that require robustness
    e.g. in formatting issues for structured responses

    This context is passed on every invocation anyway so this narrows it down to an api provider or client
    - open ai
    - llama
    - gemini
    - claude

    Within each of these providers the context can choose a model size/version
    """
    context = context or CallingContext()

    """default"""
    return GptModel()



from funkyprompt.core import AbstractEntity, typing, Field, OpenAIEmbeddingField
import datetime
from . import GenericEntityTypes
from pydantic import model_validator
from ast import literal_eval
from funkyprompt.core.utils.ids import funky_id


class Project(AbstractEntity):
    class Config:
        name: str = "project"
        namespace: str = "public"
        description: str = (
            """Projects allow people to manage what they care about, their goals etc. 
            It is possible to add and search projects and build relationships between projects and other entities"""
        )

    name: str = Field(description="The unique name of the project")
    description: str = OpenAIEmbeddingField(
        description="The detailed description of the project"
    )
    target_completion: typing.Optional[datetime.datetime] = Field(
        default=None, description="An optional target completion date for the project"
    )
    labels: typing.Optional[typing.List[str] | str] = Field(
        default_factory=list,
        description="Optional category labels - should link to topic entities",
        entity_name=GenericEntityTypes.TOPIC,
    )

    @model_validator(mode="before")
    @classmethod
    def _types(cls, values):
        """we should be stricter in array/list types but here
        example of allowing lists as TEXT in stores
        """

        if isinstance(values.get("labels"), str):
            try:
                values["labels"] = literal_eval(values["labels"])
            except:
                pass

        return values


class Task(Project):
    class Config:
        name: str = "task"
        namespace: str = "public"
        description: str = (
            """Tasks allow people to manage small objectives as part of large projects. 
            It is possible to add and search tasks and build relationships between tasks and other entities"""
        )

    project: typing.Optional[str] = Field(
        default_factory=list,
        description="The associated project",
        entity_name=GenericEntityTypes.PROJECT,
    )

    status: typing.Optional[str] = Field(
        default="TODO",
        description="The status of the project e.g. TODO, DONE",
    )

    @model_validator(mode="before")
    @classmethod
    def _ids(cls, values):
        """tasks take ids based on their project and name
        it is up to the caller to ensure uniqueness
        """
        proj = values.get("project")
        name = f"{proj}/{values['name']}"
        values["id"] = funky_id(name)
        return values

    # TODO: im testing adding the inline task - but actually the agent should know this usually if we design things right (either the agent is Task or the planner provides the metadata)
    # TODO also testing moving crud to base class so that we can assume it on a type but using its schema and not the generic one in doc strings

    @classmethod
    def add(cls, task: "Task", **kwargs):
        """Save or update a task based on its task name as key

        #task model

        ```python
        class Task(BaseModel):
            name: str
            description: str
            project: Optional[str] = None
            labels: Optional[list[str]] = []
            target_completion: Optional[datetime]
        ```

        Args:
            task: The task object to add
        """
        from funkyprompt.services import entity_store

        if isinstance(task, dict):
            task = cls(**task)

        return entity_store(cls).update_records(task)

    @classmethod
    def set_task_status(cls, task_names: typing.List[str], status: str):
        """Move all tasks by name to the given status

        Args:
            task_names (typing.List[str]): list of one or more tasks for which to change status
            status (str): status as TODO or DONE
        """
        from funkyprompt.services import entity_store

        if task_names and not isinstance(task_names, list):
            task_names = [task_names]

        q = f"""UPDATE {cls.sql().table_name} set status=%s WHERE name = ANY(%s)"""

        return entity_store(cls).execute(q, (status, task_names))

    @classmethod
    def set_task_target_completion(
        cls, task_names: typing.List[str], date: str | datetime.datetime
    ):
        """Move all tasks by name to the given status

        Args:
            task_names (typing.List[str]): list of one or more tasks for which to change status
            date (str): the new date to complete the task
        """
        from funkyprompt.services import entity_store

        if task_names and not isinstance(task_names, list):
            task_names = [task_names]

        q = f"""UPDATE {cls.sql().table_name} set target_completion=% WHERE name = ANY(%)"""

        return entity_store(cls).execute(q, (date, task_names))

    @classmethod
    def run_search(
        cls,
        questions: typing.List[str] | str,
        after_date: typing.Optional[dict] | str = None,
    ):
        """Query the tasks by natural language questions
        Args:
            questions (typing.List[str]|str): one or more questions to search for tasks
            date (str): the new date to complete the task
        """
        from funkyprompt.services import entity_store

        return entity_store(cls).ask(questions, after_date=after_date)



import typing
from pydantic import BaseModel, Field
from . import DEFAULT_MODEL

DEFAULT_MAX_AGENT_LOOPS = 10
DEFAULT_MODEL_TEMPERATURE = 0.0


class ApiCallingContext(BaseModel):
    """calling context object - all have defaults
    an agent session uses these things to control how to communicate with the user or the LLM Api
    """

    session_id: typing.Optional[str] = Field(
        default=None, description="A goal orientated session id"
    )

    session_context: typing.Optional[str] = Field(
        default=None,
        description="For routing purposes, describe the session's objective",
    )

    prefer_json: typing.Optional[bool] = Field(
        default=False, description="If the json format is preferred in response"
    )
    response_model: typing.Optional[str] = Field(
        default=None, description="A Pydantic format model to use to respond"
    )
    username: typing.Optional[str] = Field(
        default=None, description="The session username"
    )
    channel_context: typing.Optional[str] = Field(
        default=None,
        description="A channel id e.g. slack channel but more broadly any grouping",
    )
    channel_ts: typing.Optional[str] = Field(
        default=None, description="A channel conversation id e.g. slack timestamp (ts)"
    )

    prefers_streaming: typing.Optional[bool] = Field(
        default=False,
        description="Indicate if a streaming response is preferred with or without a callback",
    )
    temperature: typing.Optional[float] = Field(
        default=DEFAULT_MODEL_TEMPERATURE, description="The LLM temperature"
    )
    plan: typing.Optional[str] = Field(
        default=None,
        description="A specific plan/prompt to override default agent plan",
    )
    max_iterations: typing.Optional[int] = Field(
        default=DEFAULT_MAX_AGENT_LOOPS,
        description="Agents iterated in a loop to call functions. Set the max number of iterations",
    )
    model: typing.Optional[str] = Field(
        default=DEFAULT_MODEL, description="The LLM Model to use"
    )

    file_uris: typing.Optional[typing.List[str]] = Field(
        description="files associated with the context", default_factory=list
    )

    def get_response_format(cls):
        """"""
        if cls.prefer_json:
            return {"type": "json_object"}


class CallingContext(ApiCallingContext):
    """add the non serializable callbacks"""

    streaming_callback: typing.Optional[typing.Callable] = Field(
        default=None,
        description="A callback to stream partial results e.g. print progress",
    )
    response_callback: typing.Optional[typing.Callable] = Field(
        default=None,
        description="A callback to send final response e.g a Slack Say method",
    )

    @property
    def is_streaming(cls):
        """the streaming mode is either of these cases"""
        return cls.prefers_streaming or cls.streaming_callback is not None



import typing
from pydantic import BaseModel, Field, model_serializer, model_validator
from funkyprompt.core import AbstractModel, AbstractEntity
import datetime
import json
from funkyprompt.core.agents import CallingContext


class Message(BaseModel):
    role: str
    content: str | dict
    name: typing.Optional[str] = Field(
        description="for example function names", default=None
    )


class UserMessage(Message):
    role: str = "user"


class SystemMessage(Message):
    role: str = "system"


class MessageStack(BaseModel):
    """
    Message stack abstracts our notion of changing message context and provides some basic helpers
    Experiment with;
    - type prompts
    - known entities
    - smart pruning of context
    """

    model: AbstractModel | type = Field(
        description="The Model for the guidance of the agents - a model contains Config and functions and is required - it can be anything with a `get_model_description` method for this purpose"
    )

    current_date: typing.Optional[datetime.datetime] = Field(
        description="Id data is added to messages it can guide agents that are confused about the fourth dimension",
        default=None,
    )

    # could add base messages to but depending on what they are they could be typed things

    function_names: typing.Optional[typing.List[str]] = Field(
        description="Listing function names can be a useful hint", default_factory=list
    )

    question: typing.Optional[str] = Field(
        description="The users question", default=None
    )

    messages: typing.Optional[typing.List[Message]] = Field(
        description="A list of messages sent to the language mode context",
        default_factory=list,
    )

    language_model_provider: typing.Optional[str] = Field(
        description="The model provider can affect the serialization", default=None
    )

    known_entities: typing.Optional[typing.List[AbstractEntity]] = Field(
        default_factory=list,
        description="Known entities can be useful metadata if we encounter a small number of entities in the session - it costs context but if used can be powerful metadata",
    )

    @model_validator(mode="before")
    def _create(cls, values):
        """
        we basically want a prompt and a question but we can add in other stuff too
        """
        messages = []

        model = values.get("model")
        prompt = "Answer the users questions using world knowledge or any provided functions. Always check that you can use functions that you have. Do not use a search function if another function can be used in place. "
        if hasattr(model, "get_model_description"):
            prompt += model.get_model_description()
        """update messages from context, model and question"""

        messages.append(SystemMessage(content=prompt))
        date = values.get("current_date")
        if date:
            """this is added because sometimes it screws up date based queries"""
            messages.append(
                SystemMessage(
                    content=f"I observe the current date is {date} so I should take that into account if asked questions about time"
                )
            )

        function_names = values.get("function_names")
        if function_names:
            """this is added because sometimes it seems to need this nudge (TODO:)"""
            messages.append(
                UserMessage(
                    content=f"You can use the following functions by default {function_names} any in some cases you may be able to search and load others"
                )
            )

        """finally add the users question"""
        if values.get("question"):
            messages.append(UserMessage(content=values.get("question")))

        values["messages"] = messages

        return values

    def reset(cls):
        """returns the message stack only with the model elements"""
        data = dict(vars(cls))
        data["messages"] = []
        return MessageStack(model=cls.model, current_date=cls.current_date)

    def add(cls, message: Message | dict):
        """augment the message and retain all the other stuff"""
        if isinstance(message, dict):
            message = Message(**message)
        messages = cls.messages
        messages.append(message)
        data = dict(vars(cls))
        data["messages"] = messages

        return MessageStack(**data)

    def add_system_message(cls, data: str):
        """add string or dict content as system message"""
        return cls.add(SystemMessage(content=data))

    def add_user_message(cls, data: str | dict):
        """add string or dict content as system message"""
        return cls.add(UserMessage(content=data))

    @classmethod
    def structure_question(
        cls, question: str, model: AbstractModel, context: CallingContext = None
    ) -> typing.List[dict]:
        """Prompt building in `funkyprompt` is governed by models or types. A question is injected into a scaffold based on the type
        For language models we can construct a set of one or more messages using different roles.
        One large blob in a question can work but its often sensible to split fragments by role

        Args:
            question (str): the users question
            model (AbstractModel): the model used to generate the plan
            context (CallingContext, optional): context is used to determine the model provider and other session context. Defaults to None.
        """
        return []

    @classmethod
    def format_function_response_data(
        cls, name: str, data: typing.Any, context: CallingContext = None
    ) -> Message:
        """format the function response for the agent - essentially just a json dump

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted messages for agent as a dict
        """

        return Message(
            role="function",
            name=f"{str(name)}",
            content=json.dumps(
                {
                    # do we need to be this paranoid most of the time?? this is a good label to point later stages to the results
                    "about-these-data": "here are some data that may or may not contain the answer to your question - please review it carefully",
                    "data": data,
                },
                default=str,
            ),
        )

    @classmethod
    def format_function_response_type_error(
        cls, name: str, ex: Exception, context: CallingContext = None
    ) -> Message:
        """type errors imply the function was incorrectly called and the agent should try again

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted error messages for agent as a dict
        """
        return Message(
            role="system",
            name=f"{str(name.replace('.','_'))}",
            content=f"""You have called the function incorrectly - try again {ex}""",
        )

    def format_function_response_error(
        name: str, ex: Exception, context: CallingContext = None
    ) -> Message:
        """general errors imply something wrong with the function call

        Args:
            name (str): the name of the function
            data (typing.Any): the function response
            context (CallingContext, optional): context such as what model we are using to format the message with

        Returns: formatted error messages for agent as a dict
        """

        return Message(
            role="system",
            name=f"{str(name.replace('.','_'))}",
            content=f"""This function failed - you should try different arguments or a different function. - {ex}. 
                        If not data found you must search for another function if you can to answer the users question. 
                        Otherwise check the error and consider your input parameters """,
        )

    """smart pruning of messages"""

    @model_serializer()
    def custom_serializer(self):
        """
        we can be smarter about this
        """

        return [m.model_dump() for m in self.messages]



"""
simple app for convenience, install the [web] components
"""

from fastapi import (
    APIRouter,
    FastAPI,
    Response,
    UploadFile,
    File,
    Form,
    Depends,
    HTTPException,
)
from http import HTTPStatus
from starlette.responses import HTMLResponse
from fastapi.responses import StreamingResponse
from funkyprompt.core.agents import ApiCallingContext, Runner
import typing
from pathlib import Path


def get_custom_redoc_html(
    *,
    openapi_url: str,
    title: str,
    redoc_favicon_url: str = "https://fastapi.tiangolo.com/img/favicon.png",
    with_google_fonts: bool = True,
) -> HTMLResponse:
    """
    This is added to include the try-it-out feature in the redocly documentation at /documentation
    """

    html = f"""
    <!DOCTYPE html>
    <html>
    <head>
    <title>{title}</title>
    <!-- needed for adaptive design -->
    <meta charset="utf-8"/>
    <meta name="viewport" content="width=device-width, initial-scale=1">
    """
    if with_google_fonts:
        html += """
    <link href="https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700" rel="stylesheet">
    """
    html += f"""
        <link rel="shortcut icon" href="{redoc_favicon_url}">
        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
          body {{
            margin: 5;
            padding: 5;
          }}
        </style>
        </head>
        <body>
            <div id="redoc-container"></div>
              <script src="https://cdn.jsdelivr.net/npm/redoc@2.0.0-rc.55/bundles/redoc.standalone.min.js"> </script>
              <script src="https://cdn.jsdelivr.net/gh/wll8/redoc-try@1.4.9/dist/try.js"></script>
              <script>
                initTry({{
                openApi: `{openapi_url}`,
                  redocOptions: {{scrollYOffset: 50}},
                }})
              </script>
        </body>
        </html>
        """
    return HTMLResponse(html)


app = FastAPI(
    summary="Funkyprompt API",
    description=f"""Provides a simple shell for testing APIs with agents""",
    title="ONE",
    openapi_url=f"/openapi.json",
    docs_url=f"/docs",
    version="0.6.1",
)

api_router = APIRouter()


@app.post("/ask")
async def ask(
    question: str,
    # the response context is merged into the form data to work with files - its not posted by the client - see swagger
    response_context: ApiCallingContext = Depends(),
    files: typing.Optional[typing.List[UploadFile]] = File(..., default_factory=list),
) -> StreamingResponse:
    """
    illustrate a simple web client for the bits
    """

    try:
        """interact and respond"""
        r = Runner()
        stream = r(question=question, context=response_context)

        def iterate_s():
            """this is a bit of a hack because i cannot ensure yet that i am streaming non null things from above"""
            for s in stream:
                if s:
                    yield s

        return (
            StreamingResponse(iterate_s(), media_type="text/event-stream")
            if response_context.prefers_streaming
            else Response(content=stream)
        )

    except Exception as ex:

        raise HTTPException(status_code=500, detail=str(ex))
    
def start():
    import uvicorn

    uvicorn.run(
        f"{Path(__file__).stem}:app",
        host="0.0.0.0",
        port=5005,
        log_level="debug",
        reload=True,
    )


if __name__ == "__main__":
    """
    You can start the dev with this in the root
    uvicorn funkyprompt.app.main:app --port 5005 --reload
    
    #the swagger will be at the /funky/docs location
    #the redocly will be at the /funky/documentation location
    http://127.0.0.1:5005/funky/docs
    
    """
    
    start()
    



from . import AbstractModel
from funkyprompt.core.fields.annotations import OpenAIEmbeddingField
import typing
from pydantic import BaseModel, Field

CONVERSATION_MODEL_DESCRIPTION = f"""Conversation model tracks the ongoing conversation 
which can involve questions, answers, objectives, entities, knowledge, users, channels and more.
All of this is captured as users interact with agents
"""


class ConversationModel(AbstractModel):
    class Config:
        name: str = "conversation"
        namespace: str = "core"
        description: str = CONVERSATION_MODEL_DESCRIPTION

    user_id: str = Field(
        default="system", description="The user involved in the session"
    )
    objective_node_id: typing.Optional[str] = Field(
        description="The objective node if it exists and is being tracked", default=None
    )
    content: dict | str = OpenAIEmbeddingField(
        description="Structured or unstructured conversation - should normally take the forma of a fragment e.g. question and response"
    )



from funkyprompt.core.AbstractModel import AbstractModel, AbstractEntity, Field, typing
from funkyprompt.core.ConversationModel import ConversationModel
from funkyprompt.core.fields.annotations import OpenAIEmbeddingField


def load_entities() -> typing.List[AbstractEntity]:
    from funkyprompt.core.types.inspection import get_classes

    entities = get_classes(AbstractEntity)

    def _not_private(e):
        """we treat entities that are prefixed with _ as hidden entities"""
        return (
            e.__name__[:1] != "_"
            and not getattr(e.Config, "is_hidden", False)
            and e.__name__ != "AbstractEntity"
        )

    entities = [e for e in entities if hasattr(e, "Config")]
    entities = [e for e in entities if _not_private(e)]

    return entities


from funkyprompt.core.utils import logger



from funkyprompt.core import AbstractModel
import typing
from funkyprompt.services.data import DataServiceBase


class LanceDBService(DataServiceBase):
    """the duckdb sql model uses lancedb as an assistant vector index"""

    def create_model(self, model: AbstractModel):
        pass

    def update_records(self, records: typing.List[AbstractModel]):
        pass

    def select_one(self, id: str):
        pass

    def ask(self, question: str):
        pass



from inspect import getmembers, isfunction, isclass, ismethod, ismodule
import types
import typing
from pydantic import BaseModel, Field
from enum import Enum
import importlib, pkgutil


class TypeInfo(BaseModel):
    name: str = Field(description="parameter name")
    type: typing.Type = Field(
        description="the python type resolve from precedence of required"
    )
    args: tuple | typing.Type = Field(description="All type options in union types")
    input: typing.Any = Field(
        description="The full annotation input that we are mapping"
    )
    is_required: bool = Field(description="required or not")
    is_list: bool = Field(description="is a collection/list type or not")
    enum_options: typing.Optional[typing.List[str]] = Field(
        description="enums can provide option hints", default_factory=list
    )

    def to_json_schema(cls) -> dict:
        return {}


def get_class_and_instance_methods(cls):
    """inspect the methods on the type for methods"""
    methods = []
    class_methods = []

    for name, member in getmembers(cls):
        if isfunction(member) or ismethod(member):
            # Check if the method belongs to the class and not inherited
            if member.__qualname__.startswith(cls.__name__):
                if isinstance(member, types.FunctionType):
                    methods.append(getattr(cls, name))
                elif isinstance(member, types.MethodType):
                    class_methods.append(getattr(cls, name))

    return methods + class_methods


def resolve_signature_types(fn: typing.Callable, **kwargs):
    """given a function, resolve all signature annotations in out opinionated way"""
    return [resolve_named_type(k, v) for k, v in typing.get_type_hints(fn).items()]


def resolve_named_type(name: str, t: type, **kwargs):
    """a simple opinionated type mapping - we may pass some precedence for different modes
    the tuple that comes from `typing.get_type_hints` is passed in
    """

    def apply_precedence(t):
        """a primitive reduction - just reduce the inner most left type for now
        Example
            #this is a non required list of type string
            typing.Optional[typing.List[str | dict]]

        ultimately json stuff is sent over the wire so for now we want our functions
        to be well behaved and parse the inputs via coercian
        """
        if isinstance(t, tuple):
            t = t[0]
        args = typing.get_args(t)
        if len(args) > 0:
            t = args[0]
            return apply_precedence(t)
        return t

    args = typing.get_args(t)
    required = types.NoneType not in typing.get_args(t)
    contains_list = False
    for item in args:
        if typing.get_origin(item) in {list, typing.List}:
            contains_list = True

    """generate the summary object"""
    args = args or t
    T = apply_precedence(args)

    """handle enum types and options"""
    enum_options = []
    if isclass(T) and issubclass(T, Enum):
        enum_options = [member.value for member in T]
        if len(enum_options):
            T = type(enum_options[0])

    return TypeInfo(
        name=name,
        type=T,
        args=args,
        is_required=required,
        is_list=contains_list,
        input=t,
        enum_options=enum_options,
    )


def get_classes(
    base_filter: type = None,
    package="funkyprompt.core",
    exclusions: typing.List[str] = None,
) -> list[type]:
    """recurse and get classes implementing a base class
    we do some env stuff to make inspection work well and we should test for this e.g. dont load models and do not init singletons
    """

    exclusions = exclusions or []
    classes_in_package = []
    # Go through the modules in the package
    for _importer, module_name, is_package in pkgutil.iter_modules(
        importlib.import_module(package).__path__
    ):
        full_module_name = f"{package}.{module_name}"
        if full_module_name in exclusions:
            continue
        # Recurse through any sub-packages
        if is_package:
            classes_in_subpackage = get_classes(
                package=full_module_name, base_filter=base_filter
            )
            classes_in_package.extend(classes_in_subpackage)

        # Load the module for inspection
        module = importlib.import_module(full_module_name)

        # Iterate through all the objects in the module and
        # using the lambda, filter for class objects and only objects that exist within the module
        for _name, obj in getmembers(
            module,
            lambda member, module_name=full_module_name: isclass(member)
            and member.__module__ == module_name,
        ):
            classes_in_package.append(obj)
    return (
        classes_in_package
        if not base_filter
        else [c for c in classes_in_package if issubclass(c, base_filter)]
    )



from funkyprompt.core import AbstractModel
from funkyprompt.core.agents import (
    CallingContext,
    MessageStack,
    LanguageModel,
    Plan,
)
from ..functions import Function
import typing


class FunctionManager:
    """The function manager is used to plan and search over functions.
    It can also do the basic serialization faff to make functions in their various formats callable.
    The benefit of function managers are as follows;
    - formatting of available functions to send to LLM
    - searching and planning over functions
    - loading functions into the runtime so they can actually be called
    - generally supporting a dynamic function loading, planning and execution pattern
    """

    def __init__(self):
        """some options such as models or data stores to use for function loading"""
        self._functions = {}

    def __getitem__(self, key):
        return self._functions.get(key)

    def __setitem__(self, key, value):
        self._functions[key] = value

    def register(self, model: AbstractModel, include_function_search: bool = False):
        """register the functions of the model
        When registration is done, the functions are added to the stack of functions a runner can use

        Args:
            model (AbstractModel): a model that describes the resources and objectives of an agent
            include_function_search (bool, optional): this allows for dynamic function loading via a help command
        """
        for f in model.get_class_and_instance_methods():
            self.add_function(f)

    def add_function(self, f: typing.Callable | "Function"):
        """A callable function or Function type can be added to available functions.
        The callable is a python instance function that can be wrapped in a Function type
        or the Function can type can be added directly.
        Function types provide a reference to the actual callable function but also metadata.

        Args:
            f (typing.Callable|Function): the callable function or function descriptor to add to available functions
        """

        if not isinstance(f, Function) and callable(f):
            f = Function.from_callable(f)

        self[f.name] = f

        return f

    # may add tenacity retries for formatting
    def plan(self, question: str, context: CallingContext = None):
        """Given a question, use the known functions to construct a function calling plan (DAG)

        Args:
            question (str): any prompt/question
            context (CallingContext, optional): calling context may be used e.g. to choose the underlying model
        """

        """determine the model from context or default"""
        from funkyprompt.services.models import language_model_client_from_context

        lm_client: LanguageModel = language_model_client_from_context(context)

        """there are no functions in this context as we want a direct response from context"""
        functions = None

        # example not just of response model but add functions/prompts with the model
        """we can structure the messages from the question and typed model"""
        messages = MessageStack(
            question, model=Plan, language_model_provider=context.model
        )

        response: Plan = Plan.model_validate_json(
            lm_client(messages=messages, functions=functions, context=context)
        )

        return response

    def add_functions_by_name(self, function_names: str | typing.List[str]):
        """functions by named can be added to the runtime
        When plans or searches or run, the agent must ask to activate the functions.

        Activation means
        1. adding the function to the stack of callable functions in the language model context
        2. adding the function to the runtime so it can be called by the Runner

        Args:
            function_names (str | typing.List[str]): provide one or functions as a list
        """
        pass

    def reset_functions(self):
        """hard reset on what we know about"""
        self.functions = {}

    def search(self, question: str, limit: int = None, context: CallingContext = None):
        """search a deep function registry. The plan could be used to hold many functions in an in-memory/in-context registry.
        This as cost implications as the model must keep all functions in the context.
        On the other hand, a vector search can load functions that might be interesting but it may not be accurate or optimal

        Args:
            question (str): a query/prompt to search functions (using a vector search over function embeddings)
            limit: (int): a limit of functions returned for consideration
            context (CallingContext, optional): context may provide options
        """
        from funkyprompt.services import entity_store

        return entity_store(Function).ask(question)

    @property
    def functions(self) -> typing.Dict[str, Function]:
        """provides a map of functions"""
        return self._functions



"""
The function classes are used to provide runtime context of any callable thing
Functions can also be saved in a registry and searched by their description
In this file there are some function related helper types

1. FunctionCall: trivially wraps the name and args to a function call
2. FunctionParameter: describes a parameter name, type, and details
3. FunctionMetadata: describes a function and wraps parameters
4. Function: The main class represents a Function entity 
"""

from pydantic import Field, BaseModel, PrivateAttr, model_validator, model_serializer
from funkyprompt.core import AbstractEntity
from funkyprompt.core import types as core_types, AbstractEntity
from funkyprompt.core.types.inspection import resolve_signature_types, TypeInfo
from funkyprompt.core.fields.annotations import OpenAIEmbeddingField
from funkyprompt import LanguageModelProviders
import docstring_parser
import typing
import re

"""for example open ai does not allow some stuff like dots"""
REGEX_ALLOW_FUNCTION_NAMES: str = (
    r"[^a-zA-Z0-9_]"  # n = re.sub(regex_allow_names, "", n) if regex_allow_names else n
)
MAX_FUNCTION_NAME_LENGTH = 64
MAX_FUNCTION_DESCRIPTION_LENGTH = 1024


DESCRIPTION = f"""Functions provide an interface over resources/tools that a language model can use.
Functions can be API alls, runtime python functions, database client etc and it really does not matter which is which.
The important thing is functions provide metadata (doc strings) about how they should be used along with parameter descriptions
With this information a language model can plan over functions and invoke functions to answer user questions and solve problems.
"""


class FunctionCall(BaseModel):
    name: str
    arguments: str | dict


class FunctionParameter(BaseModel):
    """simple model for function parameters"""

    name: str
    is_required: bool = True
    description: str = Field(description="The parameter description")
    type: str = Field(description="The json schema type of the parameter")
    enum_options: typing.Optional[typing.List[str]] = Field(
        description="enums can provide option hints", default_factory=list
    )
    # format or mode if its some thing we need

    @model_serializer()
    def dump(self):
        """make more compact"""
        d = dict(vars(self))
        if not d.get("enum_options"):
            d.pop("enum_options")
        return d

    @classmethod
    def from_type_info(cls, type_info: TypeInfo, description: str):
        """the parameter descriptions are merged with annotations to produce the parameter info"""
        data = dict(vars(type_info))
        data["type"] = core_types.PYTHON_TO_JSON_MAP.get(data["type"], "object")
        if type_info.is_list:
            data["type"] = "list"

        data["description"] = f"{description}"
        return FunctionParameter(**data)

    def to_json_spec(cls, **kwargs):
        """
        dump in a json schema format ala openai
        https://cookbook.openai.com/examples/how_to_call_functions_with_chat_models
        """
        d = {"type": cls.type, "description": cls.description}
        if cls.enum_options:
            d["enums"] = cls.enum_options
        return {cls.name: d}


class FunctionMetadataParser(BaseModel):
    """simple model for function metadata - its a `Function` subset just to factor out parsing"""

    name: str = Field(description="function name")
    description: str = Field(
        description="the function description as per the docstring"
    )
    parameters: typing.List[FunctionParameter] = Field(
        default_factory=list, description="structured list of parameters"
    )

    @classmethod
    def parse_metadata(
        cls, fn: typing.Callable, alias: str = None, augment_description: str = None
    ) -> "FunctionMetadataParser":
        """
        parses some expected doc string formats and produces a description and parameter list
        the best practice of type annotations and good doc strings is assumed here

        the external lib supports some stand doc styles
        https://github.com/rr-/docstring_parser
        "Currently support ReST, Google, Numpydoc-style and Epydoc docstrings."

        Args:
            fn: the callable function
            alias: an optional alias to rename the function
            augmented_description: an optional extra context to add to the function
        """

        """for now use the external library or add custom parsers"""

        # TODO: support complex types as parameters in the most efficient way either adding them to the doc string prompt or kwargs them out in the signature

        def s_combine(*l):
            return "\n".join(i for i in l if i)

        p = docstring_parser.parse(fn.__doc__)
        description = s_combine(
            p.short_description, p.long_description, augment_description
        )
        parameter_descriptions = {p.arg_name: p.description for p in p.params}

        """function name or alias"""
        name = alias or fn.__name__
        """using type annotations get the names and types of parameters with some other info"""
        parameters = resolve_signature_types(fn)
        """map from our type-info schema which is basically the same 
           - the descriptions are merged in from doc strings"""
        parameters = [
            FunctionParameter.from_type_info(p, parameter_descriptions.get(p.name))
            for p in parameters
        ]
        return FunctionMetadataParser(
            name=name, description=description, parameters=parameters
        )


class Function(AbstractEntity):
    """a Function is a core entity in `funkyprompt`
    - map apis and databases etc as functions
    - search for functions
    - load and invoke
    - pass descriptions to language models
    """

    class Config:
        name: str = "function"
        namespace: str = "core"
        description = DESCRIPTION

    name: str = Field(description="the fully qualified function name")
    description: str = Field(
        description="the function description as in doc strings and as used to prompt"
    )
    searchable_description: typing.Optional[str] = OpenAIEmbeddingField(
        description="extended content used primarily for searching (vector search) if the docstring is not detailed enough",
        default=None,
    )
    parameters: typing.List[FunctionParameter] = Field(
        description="The parameter types and descriptions", default_factory=list
    )
    metadata: typing.Optional[dict] = Field(
        default_factory=dict,
        description="Any metadata associated with the function used in instantiation of a callable object such as security or factory providers",
    )

    @model_validator(mode="before")
    def _vals(cls, values):
        """ensure we have a description for the function for search"""
        values["extended_description"] = (
            values.get("extended_description") or values["description"]
        )
        return values

    def to_json_spec(
        cls, model_provider: str = LanguageModelProviders.openai, **kwargs
    ) -> dict:
        """dump in a json schema format ala openai
        https://cookbook.openai.com/examples/how_to_call_functions_with_chat_models
        """

        """parameters - a map of stff"""
        props = {p.name: p.to_json_spec(**kwargs) for p in cls.parameters}
        """body - we parse the name but this must marry up with what we use elsewhere"""
        name = re.sub(REGEX_ALLOW_FUNCTION_NAMES, "", cls.name)
        return {
            # "type": "function",
            # "function": {
            "name": name,
            "description": cls.description[:MAX_FUNCTION_DESCRIPTION_LENGTH],
            "parameters": {"type": "object", "properties": props},
            # },
        }

    @classmethod
    def from_library_function(
        cls,
        function_name: str,
        instance_type: str = None,
        alias: str = None,
        searchable_description: str = None,
    ) -> "_RunTimeFunction":
        """for any function that can be valuated in your library, a fully qualified name is provided to eval it

        Args:
            function_name (str): the fully qualified name of the function in your library
            instance_type (str): if the function is an instance of a type, provide the constructor
        """

        instance = core_types.eval(function_name, instance_type)
        return cls.from_callable(
            instance, alias=alias, searchable_description=searchable_description
        )

    @classmethod
    def from_callable(
        cls,
        fn: typing.Callable,
        alias: str = None,
        searchable_description: str = None,
        augment_description: str = None,
    ) -> "_RunTimeFunction":
        """given a callable, return a Function wrapper
        Alias is used because the function name e.g. if its on an instance, will not tell the whole story

        Args:
            fn (typing.Callable): any callable python object
            alias (str): an alternative name to use to name the function
        """

        """
        get the parameters and their descriptions - the kernel of the function data
        """
        function_desc = FunctionMetadataParser.parse_metadata(
            fn, alias=alias, augment_description=augment_description
        )

        """get the function description"""
        return _RunTimeFunction(
            **function_desc.model_dump(),
            searchable_description=searchable_description,
            _function=fn,
        )

    @classmethod
    def from_openapi_endpoint(
        cls, name: str, open_api_json_or_url: str | dict, bearer_token_key: str = None
    ) -> "_RunTimeFunction":
        """Given a named endpoint and the spec, provide a callable function description.
           (Maybe need some token provider beyond bearer)

        Args:
            name: endpoint name
            open_api_json_or_url (str | dict): provide schema or url to schema
            bearer_token_key: a reference to the environment loading of bearer token
        """
        pass

    """convenience callable behaviors - not serialized"""
    _function: typing.Optional[typing.Callable] = PrivateAttr(default=None)

    def __call__(self, *args, **kwargs):
        function = getattr(self, "function", None)
        if function:
            try:
                return function(*args, **kwargs)
            except:
                import traceback

                # print(traceback.format_exc())
                # undecided about bound versus unbound functions
                return function.__func__(*args, **kwargs)
        raise Exception(
            "The function object is not callable because there is no underling function object on the instance"
        )


class _RunTimeFunction(Function):
    """
    this is used in place of the one we serialize so that we can hold the function and call it
    """

    function: typing.Optional[typing.Callable] = None

    @model_validator(mode="before")
    def _init_content(cls, values):
        """
        add the function from the private one
        """
        values["function"] = values.get("_function")
        return values

    @model_serializer()
    def dump(self):
        """functions are removed"""
        d = dict(vars(self))
        if "function" in d:
            d.pop("function")

        return d



import typing

DEFAULT_EMBEDDING_MODEL = "text-embedding-ada-002"


def embed_collection(c: typing.List[str], provider="open_ai"):
    """get an embedding using the default embedding model"""
    from openai import OpenAI

    r = OpenAI().embeddings.create(input=c, model=DEFAULT_EMBEDDING_MODEL)
    return [e.embedding for e in r.data]


def embed_frame(
    data: typing.List[dict], field_mapping: dict = None, id_column: str = None
) -> typing.List[dict]:
    """given a data frame with texts that require embeddings do the thing and return a collection of records
    default conventions
    key->id
    field mappings for any column pair X and X_embedding
    """
    embeddings = {}

    def frame_col(name):
        """spoofs pandas or polars until we decide to add as lib dep"""
        return [i[name] for i in data]

    """apply conventions just for ease of testing - recommend passing via the model type"""
    if not id_column:
        id_column = "id"
    if not field_mapping:
        field_mapping = {
            k.replace("_embedding", ""): k for k in data.columns if "_embedding" in k
        }

    ids = list(frame_col(id_column))
    for field, mapping in field_mapping.items():
        text = frame_col(field)
        """use the embedding function"""
        embeddings[mapping] = embed_collection(text)

    """reshape to records"""
    keys = embeddings.keys()

    def _extract(keys, i):
        """add the embedded vectors"""
        """add the id column for updating and indexing"""
        d = {id_column: ids[i]}
        d.update({k: embeddings[k][i] for k in keys})
        return d

    return [_extract(keys, i) for i in range(len(data))]



from uuid import UUID
from funkyprompt.core.types import EMBEDDING_LENGTH_OPEN_AI
import typing
import psycopg2.extras
import uuid
from . import some_default_for_type
from typing import get_type_hints
from enum import Enum

"""special postgres attributes on pydantic fields

sql_child_relation
is_key
varchar_size
"""


"""init ensure

CREATE OR REPLACE FUNCTION update_updated_at_column()
RETURNS TRIGGER AS $$
BEGIN
    NEW.updated_at = CURRENT_TIMESTAMP;
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;

"""

"""graph

--init graph
LOAD 'age';
 SET search_path = ag_catalog, "$user", public;
SELECT create_graph('funkybrain');


create vertex

SELECT * FROM cypher('funkybrain', $$  CREATE (:label) $$) as (v agtype);


"""


class VectorSearchOperator(Enum):
    """
    If vectors are normalized to length 1 (like OpenAI embeddings), use inner product for best performance.
    see also ops
    <~> 	Hamming distance
    <%> 	Jaccard distance

    """

    L1 = " <+>"  # taxicab
    L2 = "<->"  # euclidean
    INNER_PRODUCT = "<#>"  # Neg inner product
    COSINE = "<=>"


class SqlHelper:

    def __init__(cls, model):
        from funkyprompt.core import AbstractModel

        cls.model: AbstractModel = model
        cls.table_name = cls.model.get_model_fullname()
        cls.field_names = SqlHelper.select_fields(model)
        cls.id_field = cls.model.get_model_key_field() or "id"
        cls.embedding_fields = list(cls.model.get_embedding_fields().values())
        cls.metadata = {}

    @classmethod
    def select_fields(cls, model):
        """select db relevant fields"""
        fields = []
        for k, v in model.model_fields.items():
            if v.exclude:
                continue
            attr = v.json_schema_extra or {}
            """we skip fields that are complex"""
            if attr.get("sql_child_relation"):
                continue
            fields.append(k)
        return fields

    def select_fields_with_dummies(cls):
        """selects the database fields but uses dummy values. this is to allow for some upsert modes (small hack/trick)"""

        fields = cls.select_fields(cls.model)
        model_fields = get_type_hints(cls.model)

        def dummy_value(field_name):
            ftype = model_fields[field_name]

            return some_default_for_type(ftype)

        return {f: dummy_value(f) for f in fields}

    def partial_model_tuple(cls, data: dict) -> tuple:
        """
        simple wrapper that creates a placeholder tuple injecting in partial actual data
        this is paired with partial updates
        """
        d = cls.select_fields_with_dummies()
        d.update(data)
        return tuple(d.values())

    def serialize_for_db(cls, model_instance) -> dict:
        """this exists only to allow for generalized types
        abstract models can implement db_dump to have an alt serialization path
        """

        if isinstance(model_instance, dict):
            data = model_instance
        else:
            assert isinstance(
                model_instance, cls.model
            ), f"You are trying to use a model of type {type(model_instance)} in a service of type {type(cls.model)}"
            # this is the one we want to override sometimes
            if hasattr(model_instance, "db_dump"):

                data = model_instance.db_dump()
            elif hasattr(model_instance, "model_dump"):
                assert (
                    model_instance is cls.model
                ), f"You are trying to use a model of type {type(model_instance)} in a service of type {type(cls.model)}"
                data = model_instance.model_dump()
            else:
                data = vars(model_instance)

        """if there is an embedding map we can add the embeddings here
            but its assumed that the database supports those embeddings by convention
        """

        return data

    @classmethod
    def pydantic_to_postgres_type(cls, t):
        """fill me in"""
        type_mapping = {
            str: "VARCHAR",
            int: "INTEGER",
            float: "FLOAT",
            bool: "BOOLEAN",
            dict: "JSON",
            UUID: "UUID",
            list: "ARRAY",
        }

        # TODO: need to test adding extras and other complex types like lists and json

        return type_mapping.get(t, "TEXT")

    @classmethod
    def _create_embedding_table_script(cls, entity_model, existing_columns=None):
        """for a separate embedding table
        if we have the connection we can check for a diff and create an alter statement, otherwise we must to the update
        we do not remove columns, but we can add
        """
        pass

    @classmethod
    def _create_view_script(cls, entity_model):
        """
        create or alter the view to select all columns from the join possibly with system columns
        """
        pass

    def create_script(cls, embeddings_inline: bool = True, connection=None):
        """

        (WIP) generate tables for entities -> short term we do a single table with now schema management
        then we will add basic migrations and split out the embeddings + add system fields
        we also need to add the other embedding types - if we do async process we need a metadata server
        we also assume the schema exists for now

        We will want to create embedding tables separately and add a view that joins them
        This creates a transaction of three scripts that we create for every entity
        We should add the created at and updated at system fields and maybe a deleted one

        - key register trigger -> upsert into type-name -> on-conflict do nothing

        - we can check existing columns and use an alter to add new ones if the table exists

        """
        entity_model = cls.model

        def is_optional(field):
            return typing.get_origin(field) is typing.Union and type(
                None
            ) in typing.get_args(field)

        # assert config has the stuff we need

        table_name = (
            f"{entity_model.get_model_namespace()}.{entity_model.get_model_name()}"
        )
        fields = typing.get_type_hints(entity_model)
        field_descriptions = entity_model.model_fields
        id_field = "id"  # <- the id is a hash of the name or a unique id and we use it as the constrain by convention

        columns = []
        for field_name, field_type in fields.items():
            """handle uuid option"""
            if typing.get_origin(
                field_type
            ) is typing.Union and UUID in typing.get_args(field_type):
                postgres_type = "UUID"
            else:
                postgres_type = SqlHelper.pydantic_to_postgres_type(field_type)

            field_desc = field_descriptions[field_name]
            column_definition = f"{field_name} {postgres_type}"
            # we could have a default thing but hold
            # if field_desc.field_info.default is not None:
            #     column_definition += f" DEFAULT {json.dumps(field_desc.field_info.default)}"
            if field_name == id_field:
                column_definition += " PRIMARY KEY "
            elif not is_optional(field_type):
                column_definition += " NOT NULL"
            columns.append(column_definition)

            """check should add embedding vector for any columns"""
            metadata = field_descriptions.get(field_name)
            extras = getattr(metadata, "json_schema_extra", {}) or {}
            if extras.get("embedding_provider", "").replace("_", "") == "openai":
                columns.append(
                    f"{field_name}_embedding vector({EMBEDDING_LENGTH_OPEN_AI}) NULL"
                )

            """add system fields - created at and updated at fields"""
            # TODO

        """add system fields"""
        columns.append("created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP")
        columns.append("updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP")
        columns.append("deleted_at TIMESTAMP")

        columns_str = ",\n    ".join(columns)
        create_table_script = f"""
        CREATE TABLE {table_name} (
            {columns_str}
        );
        
        CREATE TRIGGER update_updated_at_trigger
        BEFORE UPDATE ON {table_name}
        FOR EACH ROW
        EXECUTE FUNCTION update_updated_at_column();

        """
        return create_table_script

    def upsert_query(
        cls,
        batch_size: int,
        returning="*",  # ID, * etc.
        restricted_update_fields: str = None,
        # records: typing.List[typing.Any],
        # TODO return * or just id for performance
    ):
        """upserts on the ID conflict

        if deleted at set generate another query to set deleted dates for records not in the id list

        This will return a batch statement for some placeholder size. You can then

        ```
        connector.run_update(upsert_sql(...), batch_data)

        ```

        where batch data is some collection of items

        ```
        batch_data = [
            {"id": 1, "name": "Sample1", "description": "A sample description 1", "value": 10.5},
            {"id": 2, "name": "Sample2", "description": "A sample description 2", "value": 20.5},
            {"id": 3, "name": "Sample3", "description": "A sample description 3", "value": 30.5},
        ]
        ```
        """

        """TODO: the return can be efficient * for example pulls back embeddings which is almost never what you want"""
        field_list = cls.field_names
        """conventionally add in order anything that is added in upsert and missing"""
        for c in restricted_update_fields or []:
            if c not in field_list:
                field_list.append(c)

        non_id_fields = [f for f in field_list if f != cls.id_field]
        insert_columns = ", ".join(field_list)
        insert_values = ", ".join([f"%({field})s" for field in field_list])

        """restricted updated fields are powerful for updates 
           we can ignore the other columns in the inserts and added place holder values in the update
        """
        update_set = ", ".join(
            [
                f"{field} = EXCLUDED.{field}"
                for field in restricted_update_fields or non_id_fields
            ]
        )

        value_placeholders = ", ".join(
            [f"({insert_values})" for _ in range(batch_size)]
        )

        value_placeholders = "%s"

        """batch insert with conflict - prefix with a delete statement that sets items to deleted"""
        upsert_statement = f"""
        -- now insert
        INSERT INTO {cls.table_name} ({insert_columns})
        VALUES {value_placeholders}
        ON CONFLICT ({cls.id_field}) DO UPDATE
        SET {update_set}
        RETURNING {returning};
        """

        return upsert_statement.strip()

    def embedding_fields_partial_update_query(
        cls, batch_size: int, returning: str = "*"
    ):
        """for now using a convention but this should be determined from the model
        we have added a convention on the restricted fields for now to reuse the partial update
        in this mode we are adding embedding fields that are not on the model schema but the convention is to add the embedding fields in order
        later we should be more explicit about this
        a decision is made to treat the embedding just like a hidden index that is not on the model - therefore it lives on the client/infra only and not the model
        we build this into the SQL adapters and postgres client
        """

        """notice we apply a convention for embedding fields"""
        return cls.partial_update_query(
            field_names=[f for f in cls.embedding_fields],
            batch_size=batch_size,
            returning=returning,
        )

    def partial_update_query(cls, field_names, batch_size: int, returning: str = "*"):
        """
        this is just a slight mod on the other one - we could refactor to just have a field restriction
        """

        return cls.upsert_query(
            batch_size=batch_size,
            returning=returning,
            restricted_update_fields=field_names,
        )

    def query_from_natural_language(
        self,
        question: str,
    ):
        """"""
        pass



# https://github.com/groq/groq-python

from groq import Groq

# https://console.groq.com/docs/models
model = "llama3-8b-8192"

# SET: GROQ_API_KEY
# client = Groq()

# model = "llama3-8b-8192"

# chat_completion = client.chat.completions.create(
#     messages=[
#         {
#             "role": "user",
#             "content": "Explain the importance of low latency LLMs",
#         }
#     ],
#     model="llama3-8b-8192",
# )
# print(chat_completion.choices[0].message.content)



POSTGRES_DB = "postgres"
POSTGRES_SERVER = "localhost"
POSTGRES_PORT = 5432
POSTGRES_PASSWORD = "password"
POSTGRES_USER = "sirsh"
POSTGRES_CONNECTION_STRING = f"postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@{POSTGRES_SERVER}:{POSTGRES_PORT}/{POSTGRES_DB}"
AGE_GRAPH = "funkybrain"



"""
The runner can call the LLM in a loop and manage the stack of messages and functions
Function calling and streaming is handled
Open telemetry is used to publish metrics which a collector could manage
If this goes above 200 lines of codes we have failed!
"""

from funkyprompt.core import AbstractModel
from funkyprompt.services.models import language_model_client_from_context
from funkyprompt.core import utils
from funkyprompt.core.agents import (
    CallingContext,
    DefaultAgentCore,
    LanguageModel,
)

from funkyprompt.core import ConversationModel
from funkyprompt.services import entity_store
from . import MessageStack
from . import FunctionCall, FunctionManager, Function
import typing


class Runner:
    """Runners are simple objects that provide the interface between types and language models
    The message setup is the only function that plays with natural language.
    While almost all of the "prompting" is pushed out to types and functions,
    This setup function is the one function you can play with to make sure the comms are right with the LLM.
    For example it is here we inject plans and questions and other hints for how to run things.
    But by design, the critical guidance should be abstracted by Types and Functions.
    Beyond this, the rest is routine;
    - import type metadata and functions from the model which controls most everything
    - run an executor loop sending context to the LLM
    - implement the invocation and message setup methods to manage the function and message stack

    Under the hood the function manager handles actual function loading and searching
    """

    def __init__(self, model: AbstractModel = None):
        """
        A model is passed in or the default is used
        The reason why this is passed in is to supply a minimal set of functions
        If the model has no functions simple Q&A can still be exchanged with LLMs.
        More general the model can provide a structured response format.
        This is powerful because the model is a Pydantic annotated type but can be realized as a json response
        """
        self.model = model or DefaultAgentCore()
        self._function_manager = FunctionManager()
        self.initialize()

    def initialize(self):
        """register the functions and other metadata from the model"""

        self._context = None
        """register the model's functions which can include function search"""
        self._function_manager.register(self.model)
        self._function_manager.add_function(self.help)

    def help(self, questions: str | typing.List[str]):
        """if you are stuck ask for help with very detailed questions to help the planner find resources for you.

        Args:
            question (str): provide detailed questions to guide the planner
        """

        utils.logger.debug(f"help/{questions=}")

        try:
            plan = self._function_manager.plan(questions)
        except:
            return {"message": "planning pending - i suggest you use world knowledge"}

        """describe the plan context e.g. its a plan but you need to request the functions and do the thing -> update message stack"""

        return plan

    def invoke(self, function_call: FunctionCall):
        """Invoke function(s) and parse results into messages

        Args:
            function_call (FunctionCall): the payload send from an LLM to call a function
        """
        f = self._function_manager[function_call.name]

        try:
            """try call the function - assumes its some sort of json thing that comes back"""
            data = f(**function_call.arguments) or {}
            data = MessageStack.format_function_response_data(
                function_call.name, data, self._context
            )
            """if there is an error, how you format the message matters - some generic ones are added
            its important to make sure the format coincides with the language model being used in context
            """
        except TypeError as tex:
            utils.logger.warning(f"Error calling function {tex}")
            data = MessageStack.format_function_response_type_error(
                function_call.name, tex, self._context
            )
        except Exception as ex:
            utils.logger.warning(f"Error calling function {ex}")
            data = MessageStack.format_function_response_error(
                function_call.name, ex, self._context
            )

        """update messages with data if we can or add error messages to notify the language model"""
        self.messages.add(data)

    @property
    def functions(self) -> typing.Dict[str, Function]:
        """provide access to the function manager's functions"""
        return self._function_manager.functions

    def run(self, question: str, context: CallingContext, limit: int = None):
        """
        Ask a question to kick of the agent loop
        """

        """setup all the bits before running the loop"""
        lm_client: LanguageModel = language_model_client_from_context(context)
        self._context = context
        self.messages = MessageStack(
            model=self.model,
            question=question,
            current_date=utils.dates.now(),
            function_names=self.functions.keys(),
            language_model_provider=context.model,
        )

        """run the agent loop to completion"""
        for _ in range(limit or context.max_iterations):
            response = None
            function_descriptions = [f.to_json_spec() for f in self.functions.values()]
            """call the model with messages and function + our system context"""
            response = lm_client(
                messages=self.messages.model_dump(),
                context=context,
                functions=function_descriptions,
            )
            if isinstance(response, FunctionCall):
                """call one or more functions and update messages"""
                self.invoke(response)
                continue
            if response is not None:
                # marks the fact that we have unfinished business
                break

        """fire telemetry"""

        """log questions to store unless disabled"""

        self.dump(question, response, context)

        return response

    def dump(self, questions: str, response: str, context: CallingContext):
        """dumps the messages and context to stores
        if the session is a typed objective this is updated in a slowly changing dimension
        generally audit all transactions unless disabled
        """
        from uuid import uuid4

        default_id = str(uuid4())

        """for any pydantic response"""
        if hasattr(response, "model_dump_json"):
            response = response.model_dump_json()

        entity_store(ConversationModel).update_records(
            ConversationModel(
                id=default_id,
                user_id=context.username or "system",
                objective_node_id=context.session_id,
                content={"question": questions, response: response},
            )
        )

    def __call__(
        self, question: str, context: CallingContext = None, limit: int = None
    ):
        """
        Ask a question to kick of the agent loop
        """
        context = context or CallingContext()
        return self.run(question, context, limit=limit)


"""

#TODO: general test e.g. call a function with limit=1, it doesnt loop and confirm the result and therefore we might not get behviouar expected
generally we want to ensure consistent output
        
"""



"""generic types are sample everyday objects
-- project
-- person
-- org
-- task
-- book
-- topic
-- idea/scrapbook

each of these can define [relations] to other things
and we can also generate "aggregates" as separate concepts e.g. daily snapshots -> could go to duck or postgres

"""

from enum import Enum


class GenericEntityTypes(Enum):
    TOPIC = "topic"
    PROJECT = "project"


from funkyprompt.entities.nodes import *


from funkyprompt.core.types.inspection import get_classes
from funkyprompt.core import AbstractEntity
from funkyprompt.core import load_entities as load_core_entities


def load_entities(include_core: bool = True) -> typing.List[AbstractEntity]:
    """
    Load entities including the core ones optionally
    if funkyprompt is used as a library we can register entities from
    - registering/importing a package of entities
    - dynamic entities in a database
    """
    entities = get_classes(AbstractEntity, package="funkyprompt.entities")
    """
    add core entities
    """
    if include_core:
        entities += load_core_entities()
    return entities


def resolve(
    entity_model_name: str, entity_model_namespace: str = "public", **kwargs
) -> AbstractEntity:
    """resolve entity given the name and namespace
    the default is public for generic type entities that we store in funkyprompt
    """
    match = f"{entity_model_namespace}.{entity_model_name}"
    for e in load_entities():
        if e.get_model_fullname() == match:
            return e



from funkyprompt.core import AbstractModel, Field
import typing

DESCRIPTION = f"""The query classifier is used to determine the nature of a natural language query.
1. If entities are mentioned, then it could be a key look therefore extract entities in the question is important
2. If the question consists of multiple orthogonal elements, it could be split into separate questions
3. if the user asks for lists or stats then they probably want to run an SQL type or Tabula type of query
4. if the user asks for something more descriptive then a vector search may be most useful
5. In some complex cases a (cypher) graph query might be the most useful IF the schema of the graph is know

*How to classify*
- It is useful to emphasize entities if they exist as a high value query input. 
  - Broad topics such as 'programming' or 'cities' are not considered as entities, entities should be specific  like Order123 or New York
- if you are confident about an SQL query this could also yield a high value result for the user
    - confidence calibration is important. If there is any table, column or value that you have 
    guessed and not seen in the entity mode your confidence should be less that 0.75
    - you should NEVER use the user's question as evidence for the existence of a table or field since they have no knowledge of the data schema
- Otherwise, mention if the question require vector search because the text is to vague or you do not have enough domain semantic context
  - it is useful to decompose questions for multiple vector searches can be run
- Graph (cypher) queries are used for some complex edge vases

"""

# Notes: a big part of this is lowering the confidence on SQL


class QueryClassifier(AbstractModel):
    class Config:
        name: str = "query_classifier"
        namespace: str = "core"
        description = DESCRIPTION
        MIN_QUERY_CONFIDENCE = 0.84

    sql_query: typing.Optional[str] = Field(
        description="Provide a postgres dialect SQL for tables that match entities you know about only. The users question is not evidence of existence.",
        default_factory=list,
    )
    sql_query_confidence_based_on_model: typing.Optional[float] = Field(
        description="You should NEVER use table names, fields or predicate values that you have no evidence for in the Entity Model provided. The user's question is not evidence. Low scores are < 0.85",
        default=0.0,
    )
    confidence_comments: typing.Optional[str] = Field(
        description="explain reasons for lack of confidence.",
        default_factory=list,
    )
    cypher_query: typing.Optional[dict] = Field(
        description="Provide a cypher query AND your confidence in the query. Low confidence queries i.e. those that lack evidence in input schema will be ignore. scored as a percentage",
        default_factory=dict,
    )
    entities: typing.List[str] = Field(
        description="List of entities in the input question. Entities should NOT be general nouns or topics but specific identifiers or places or people with identity",
        default=None,
    )
    decomposed_questions: typing.List[str] = Field(
        description="The users question(s) decomposed into one or more questions"
    )

    entity_model_type: typing.Optional[str] = Field(
        default=None, description="The fully qualified model entity type if known"
    )
    recommend_query_type: str = Field(
        description="VECTOR|SQL|ENTITY|GRAPH", default=None
    )

    @classmethod
    def _classify_question_as_query_for_model(
        cls,
        model: AbstractModel,
        question: str,
        preview: bool = False,
        return_model: bool = True,
    ) -> "QueryClassifier":
        """Used to classify a question based on a model semantics.
        use the preview flag to see the prompt markdown
        use the return_model=False to have the model be more verbose in testing (see its thought process)

        Args:
            model: The model type providing semantics
            question: the user's question tp classify
            preview: this shows the generated prompt (best viewed as Markdown)
            return_model: the default returns the pydantic object for the classification and asks the language model for JSON
        """

        from funkyprompt.services.models import language_model_client_from_context
        from funkyprompt.core.agents import LanguageModel, CallingContext

        plan = f"""Below you are given details about a model. 
This semantic information corresponds to a table 
with the given fields some of which can be searched with vector search. 
Entity lookup is possible for the name field if known. 
You should use the Query Classifier Model below [A] 
to provide a classification for the *question*
in the context the model [B]

# User's Question
```
{question}
```

# [A] Query Classifier model (use this to reply in succinct Json)

{cls.get_model_as_prompt()}

# [B] Entity model

{model.get_model_as_prompt()}

"""

        if preview:
            return plan

        lm_client: LanguageModel = language_model_client_from_context()

        data = lm_client(plan, context=CallingContext(prefer_json=return_model))

        if return_model:
            return cls.model_validate_json(data)
        return data



# https://cookbook.openai.com/examples/how_to_call_functions_with_chat_models


import openai
import typing
from funkyprompt.core.agents import CallingContext
from funkyprompt.core.functions import FunctionCall
import json
from . import LanguageModelBase


def _get_function_call_or_stream(
    response, callback=None, response_buffer=None, token_callback_action=None
):
    """
    This is a little bit opaque as it tries to abstract
    function calling and non-function calling for both streaming modes
    """

    def function_builder(function_call, response):
        """wrap it in our simple interface"""
        function_name = function_call.name
        function_args = function_call.arguments
        for c in response:
            if c.choices:
                c = c.choices[0].delta
                if c.function_call:
                    function_args += c.function_call.arguments
        return FunctionCall(name=function_name, arguments=json.loads(function_args))

    def content_builder(content, response):
        """
        isolate the code path where we return a generator
        """
        yield content
        for c in response:
            if not c.choices:
                """unfortunately the streaming model is a bit tricky to get the token usage out"""
                ###############################
                try:
                    token_callback_action(c)
                except:
                    raise
                ###############################
            if c.choices:
                c = c.choices[0].delta
                yield c.content

    if isinstance(response, openai.Stream):
        """this just checks the different states of messages"""
        content = ""
        for chunk in response:

            if chunk.choices:
                _chunk = chunk.choices[0].delta
                has_call = _chunk.function_call
                if has_call:
                    fb = function_builder(
                        _chunk.function_call,
                        response,
                    )
                    return fb
                elif _chunk.content:
                    """here we optionally stream to the callback"""
                    content += _chunk.content
                    if callback:
                        callback(_chunk.content)
                    if isinstance(response_buffer, list):
                        response_buffer.append(_chunk.content)
                    # we go into generator mode when there is no callback and its a stream
                    else:
                        return content_builder(_chunk.content, response)

            elif token_callback_action:
                """how we count are pennies when streaming"""
                token_callback_action(chunk)

        if isinstance(response_buffer, list):
            response_buffer.append(content)
        return content

    else:
        """not streaming mode - respecting the same interface"""
        response_message = response.choices[0].message
        function_call = response_message.function_call
        if function_call:
            return FunctionCall(
                name=function_call.name,
                arguments=json.loads(function_call.arguments),
            )
        if isinstance(response_buffer, list):
            response_buffer.append(response_message.content)

        """not streaming mode token callback"""
        if token_callback_action:
            try:
                token_callback_action(response)
            except:
                raise

        return response_message.content


class GptModel(LanguageModelBase):

    def get_function_call_or_stream(
        self,
        response: typing.Any,
        callback: typing.Optional[typing.Callable] = None,
        response_buffer: typing.List[typing.Any] = None,
        token_callback_action: typing.Optional[typing.Callable] = None,
    ):
        return _get_function_call_or_stream(
            response=response,
            callback=callback,
            response_buffer=response_buffer,
            token_callback_action=token_callback_action,
        )

    def run(
        cls,
        messages: typing.List[dict],
        context: CallingContext,
        functions: typing.Optional[dict] = None,
        **kwargs
    ):
        """The run entry point for the agent model
        - calls the api
        - manages streaming
        - manages function calling
        """

        response = openai.chat.completions.create(
            model=context.model,
            functions=(list(functions) if functions else None),
            function_call="auto" if functions else None,
            messages=messages,
            temperature=context.temperature,
            response_format=context.get_response_format(),
            stream=context.is_streaming,
            stream_options=({"include_usage": True} if context.is_streaming else None),
        )

        cls.response_buffer = []
        response = _get_function_call_or_stream(
            response,
            context.streaming_callback,
            response_buffer=cls.response_buffer,
            # token_callback_action=None,
        )

        return response



DEFAULT_MODEL = "gpt-4o-2024-05-13"
GPT_MINI = "gpt-4o-mini"


from funkyprompt.core.functions.Function import Function, FunctionCall
from .CallingContext import CallingContext, ApiCallingContext
from .DefaultAgentCore import DefaultAgentCore
from .AbstractLanguageModel import LanguageModel
from .MessageStack import MessageStack
from .Plan import Plan
from .FunctionManager import FunctionManager
from .Runner import Runner
from .QueryClassifier import QueryClassifier



"""
simple cli
"""

import typer
import funkyprompt
from funkyprompt.core.agents import CallingContext
import typing

app = typer.Typer()


@app.command("ask")
def ask(query: typing.Optional[str] = typer.Option(None, "--query", "-q")):
    """
    run a query
    """

    # todo proper loader by name - this assumes default namespace and instruct embedding
    def callback(s):
        print(s, end="")

    query = query = "tell the user welcome to funkyprompt - run ask -q your question"
    funkyprompt.ask(query, context=CallingContext(response_callback=callback))



from pydantic import BaseModel, create_model, Field, model_validator
import uuid
import typing
from funkyprompt.core.types import inspection
from funkyprompt.core.types.sql import SqlHelper
from funkyprompt.core.types.cypher import CypherHelper
from funkyprompt.core.utils.ids import funky_id

"""
names are always unique in funkyprompt for key-value entity lookups
however there are times when the name is not unique and if so the id should be generated by the system
"""
KEY_FIELD_ATTRIBUTE = "is_key"
DEFAULT_KEY_ATTRIBUTE_NAME = "name"
DEFAULT_NAMESPACE = "default"


class AbstractModel(BaseModel):
    class Config:
        name: str = "abstract_model"
        namespace: str = "core"
        description: str = "Provide a rich model description"

    id: typing.Optional[str | uuid.UUID] = Field(
        default=None,
        description="A unique hash/uuid for the entity. The name can be hashed if its marked as the key",
    )

    @classmethod
    def get_model_name(cls):
        c = getattr(cls, "Config", None)
        if c and getattr(c, "name", None):
            return c.name
        """else infer from lib"""
        s = cls.model_json_schema(by_alias=False)
        return s.get("title", cls.__name__)

    @classmethod
    def get_model_namespace(cls):
        c = getattr(cls, "Config", None)
        if c and getattr(c, "namespace", None):
            return c.namespace
        """else infer from lib"""
        # convention
        namespace = cls.__module__.split(".")[-1]
        return (
            namespace
            if namespace not in ["model", "__main__", "entity"]
            else DEFAULT_NAMESPACE
        )

    @classmethod
    def get_model_description(cls):
        """the description of the entity - import for prompting"""
        c = getattr(cls, "Config", None)
        if c and getattr(c, "description", None):
            return c.description

    @classmethod
    def get_model_fullname(cls):
        """
        the model name is our convention e.g. meta.bodies
        usually we determine this from either a config object or the type itself
        """
        return f"{cls.get_model_namespace()}.{cls.get_model_name()}"

    @classmethod
    def get_type_fullname(cls):
        """
        the object name
        """
        return f"{cls.__module__}.{cls.__name__}"

    @classmethod
    def get_model_key_field(cls):
        """
        the field that is used as the primary key if it exists
        - otherwise a unique id should be generated by the system
        all models have ids and some of additional friendly names
        """
        s = cls.model_json_schema(by_alias=False)
        key_props = [
            k for k, v in s["properties"].items() if v.get(KEY_FIELD_ATTRIBUTE)
        ]
        if len(key_props):
            return key_props[0]

    def get_key_value(cls):
        """
        return the instance key value based on the configured key attribute name
        """
        return getattr(cls, cls.get_model_key_field())

    @classmethod
    def create_model(cls, name: str, namespace: str = None, **fields):
        """
        For dynamic creation of models for the type systems
        create something that inherits from the class and add any extra fields
        """
        namespace = namespace or cls.get_model_namespace()
        return create_model(name, **fields, __module__=namespace, __base__=cls)

    # def get_dynamic_functions

    def get_dummy_values(cls):
        """dummy values useful in some automation"""
        pass

    def to_arrow(cls):
        """convert the object to its pyarrow representation"""
        pass

    @classmethod
    def sql(cls) -> SqlHelper:
        """reference the sql helper"""

        return SqlHelper(cls)

    @classmethod
    def cypher(cls) -> CypherHelper:
        """reference the cypher helper"""

        return CypherHelper(cls)

    def db_dump(self):
        """serialize complex types as we need for DBs/Postgres
        - we do things like allow for config to turn fields off
        - we map complex types to json
        - embedding are added async on a new table in our model

        """
        from funkyprompt.core.types.sql import SqlHelper
        import json

        data = vars(self)
        """control selectable fields by exclude or other attributes"""
        fields = SqlHelper.select_fields(self)

        def check_complex(v):
            if isinstance(v, dict) or isinstance(v, list):
                return json.dumps(v)
            return v

        data = {k: check_complex(v) for k, v in data.items() if k in fields}

        return data

    @classmethod
    def get_embedding_fields(cls) -> typing.Dict[str, str]:
        """returns the fields that have embeddings based on the attribute - uses our convention"""
        needs_embeddings = {}
        for k, v in cls.model_fields.items():
            extras = getattr(v, "json_schema_extra", {}) or {}
            if extras.get("embedding_provider"):
                needs_embeddings[k] = f"{k}_embedding"
        return needs_embeddings

    @classmethod
    def get_model_as_prompt(cls) -> str:
        """the model as prompt provides a schema and also the description of the model
        if the base class implements `_get_prompting_data` then data will be loaded into context
        For example this is used in function planning

        this is experimental and may not be a perfect abstraction
        """
        injected_data = ""
        if getattr(cls, "_get_prompting_data", None) is not None:
            injected_data = cls._get_prompting_data()

        from funkyprompt.core.types.pydantic import get_pydantic_properties_string

        return f"""## {cls.get_model_fullname()}
    
    *Description*: {cls.get_model_description()}
    
    *Field info*:
    {get_pydantic_properties_string(cls)}
    
    {injected_data}
    """

    #############
    ##   INSPECTION
    #############

    @classmethod
    def get_class_and_instance_methods(cls):
        """returns the methods on the type that we care about"""

        # TODO: by default we do not show all base methods as agent-callable but we can register run_search and add here (crud)

        methods = inspection.get_class_and_instance_methods(cls)

        # for additional in ["run_search", "add"]:
        #     if hasattr(cls, additional):
        #         methods.append(getattr(cls, additional))
        """return everything but hide privates"""
        return [m for m in methods if not m.__name__[:1] == "_"]

    """
    ----------
    """

    @classmethod
    def _register(cls):
        """a not to be abused but convenient self-register in the core entity store"""
        from funkyprompt.services import entity_store

        return entity_store(cls)._create_model()


class AbstractEntity(AbstractModel):
    """the abstract entity is a sub class of model that admits a unique name0
    - abstract entities are treated as graph nodes unless the config specifies exclude_from_graph=True

    """

    name: str = Field(description="The name is unique for the entity", is_key=True)

    @model_validator(mode="before")
    @classmethod
    def _id(cls, values):
        """"""
        if not values.get("id"):
            values["id"] = funky_id(values["name"])
        return values

    @classmethod
    def run_search(
        cls, questions: str | typing.List[str], limit: int = None, **kwargs
    ) -> typing.List[AbstractModel]:
        """search the entity using the default store

        Args:
            questions (str | typing.List[str]): ask one or more questions - the more the better
            limit (int, optional): provide an optional search limit. Defaults to None.
        """

        from funkyprompt.services import entity_store

        return entity_store(cls).ask(questions, limit, **kwargs)


class AbstractContentModel(AbstractEntity):
    """use to generate generic content types
    This is useful because we can create generic types to save in the data stores in their own tables
    These have content blobs that can be saved with the open ai embedding

    Example

        ```python
        my_generic =  AbstractContentModel.create_model(name='test', namespace='public')
        #creates a new table public.test with the content model schema
        my_generic._register()
        # you can now create these objects and insert them
        ```
    """

    content: str = Field(
        description="The name is unique for the entity", embedding_provider="openai"
    )
    category: typing.Optional[str] = Field(
        default=None, description="the grouping category for the content"
    )


class AbstractImageContentModel(AbstractContentModel):
    """like the abstract content model but for image data"""

    content: str = Field(description="Image uri", embedding_provider="clip")
    description: typing.Optional[str] = Field(
        default=None,
        description="optional image description",  # TODO: may add an open embedding to this
    )



from funkyprompt.services.models import language_model_client_from_context
from funkyprompt.core import AbstractEntity
from .data.postgres import PostgresService


def entity_store(model: AbstractEntity):
    """returns the configured store for the entity"""
    return PostgresService(model)



import anthropic

# https://docs.anthropic.com/en/docs/quickstart
model = "claude-3-5-sonnet-20240620"

"""
client = anthropic.Anthropic()

message = client.messages.create(
    model="claude-3-5-sonnet-20240620",
    max_tokens=1000,
    temperature=0,
    system="You are a world-class poet. Respond only with short poems.",
    messages=[
        {
            "role": "user",
            "content": [
                {
                    "type": "text",
                    "text": "Why is the ocean salty?"
                }
            ]
        }
    ]
)
print(message.content)"""



"""this example illustrates how "agents" are just objects with properties and functions
  some agents may not have properties 
  but properties act as a response template which will be a json response.
"""

from funkyprompt.core import AbstractModel
import typing


AGENT_CORE_DESCRIPTION = """
As a funkyprompt agent you are responsible for calling 
provided functions to answer the users question.
The default agent core contains basic bootstrapping functions. 

An example use case would be to ask questions about 
named entities which can be loaded from the store. 
Once loaded, these entities provide not only details 
but references to other functions that can be called. 
This can be used to allows agent workflows to multi-hop.

Furthermore, a help function can be used for general planning 
over all known functions in a function registry.
These functions are loaded on demand into the runners for use by LLMs.

Image description functions points to some multimodal applications.

`Funkyprompt` uses the following key principles;
- it treats the dynamic parts of agents i.e. the runner as a simple shell (<200 lines of code)
- the runner outsources two stateful jobs i. the message stack (simple list of dicts) and ii. the function stack
- it then treats agents as "declarative" using a object orientated generation paradigm
- objects contains methods and fields as well as metadata. these are the agent prompts/guidance
- objects also provide a response schema if relevant to guide json agent response formats 
   - Note: the implies the response format, functions, goals and general "prompting" are all encapsulated in a single Pydantic object
- functions can be infinitely searched and the function stack can be dynamical managed in context

By treating agents as simple object types which provide rich semantics and access to encapsulated functions
`funkyprompt` allows for complex agent systems to be built in a lightweight and intuitive way

"""


class DefaultAgentCore(AbstractModel):
    """Agents in `funkyprompt` are declarative things.
    They do not do anything except expose metadata and functions.
    Runners are used to manage the comms with LLMS
    and the basic workflow - there is only one workflow in `funkyprompt`.
    This default type for use in the runner - contains basic functions.
    This minimal agent is quite powerful because it can bootstrap RAG/search.
    """

    # ideas
    # it may be that not providing a format results in default plain text / markdown

    class Config:
        name: str = "agent"
        namespace: str = "core"
        description: str = AGENT_CORE_DESCRIPTION

    @classmethod
    def describe_images(self, images: typing.List[str], question: str = None) -> dict:
        """describe a set of using the default LLM and an optional prompt/question

        Args:
            images (typing.List[str]): the images in uri format or Pil Image format
            question (str): the question to ask about the images - optional, a default prompt will be used
        """
        pass

    @classmethod
    def lookup_entity(self, keys: str | typing.List[str]) -> typing.List[dict]:
        """Given one or more entity keys, lookup the entity details

        Args:
            keys (str | typing.List[str]): a list of one or more keys

        Returns: a list of typed entities
        """
        pass

    @classmethod
    def funky_prompt_codebase(self, questions: str):
        """ask questions about the codebase aka library

        Args:
            questions (str): provide one or more questions to ask
        """
        from funkyprompt.core import load_entities

        print(f"funky_prompt_codebase/{questions=}")

        return {"questions": questions, "the following entities exist": load_entities()}



"""
Functions can be decorated for aiding resolution at runtime for LLMs
This supports stubbing patterns
"""

from pydantic import Field, BaseModel
from functools import partial
import typing

"""
[A] Provide some common shorthands for field annotations used with pydantic objects
"""


class SqlTypeFields(BaseModel):
    """things like varchar lengths"""

    varchar_length: typing.Optional[int] = Field(
        default=None,
        description="If a varchar length is used the field will be VARCHAR otherwise we might default to TEXT",
    )
    numeric_type: typing.Optional[tuple] = Field(
        default=None,
        description="Numeric types can be defined",
    )


def KeyField():
    """it is common to specify which field is the key"""
    return partial(Field, is_key=True)


def OpenAIEmbeddingField():
    """it is common to have text content or image content that can be embedded - openai will use system defaults"""
    return partial(Field, embedding_provider="openai")


def CLIPEmbeddingField():
    """it is common to have text content or image content that can be embedded - clip will use system defaults"""
    return partial(Field, embedding_provider="clip")


"""
By partially invoking we recover the doc string for the Fields that can be used as normal
"""
KeyField = KeyField()
OpenAIEmbeddingField = OpenAIEmbeddingField()
CLIPEmbeddingField = CLIPEmbeddingField()


class Example(BaseModel):
    """
    Here is an example annotation to use for reference and tests
    """

    id: str = KeyField(description="test")
    text: str = OpenAIEmbeddingField(description="example embedded property")


"""
[B] Provide some helpers to inspect type annotations. Uses include
- function signatures
- mapping pydantic objects to other formats

In some cases we apply conventions for funkyprompt that would not apply in general
"""



# https://ai.google.dev/gemini-api/docs/get-started/tutorial?lang=python

# https://ai.google.dev/gemini-api/docs/function-calling/tutorial?lang=python

# gemini-1.5-flash: our fastest multi-modal model
# gemini-1.5-pro: our most capable and intelligent multi-modal model

import google.generativeai as genai

# model = genai.GenerativeModel('gemini-1.5-flash')
# response = model.generate_content("What is the meaning of life?", stream=True)

# messages = [
#     {'role':'user',
#      'parts': ["Briefly explain how a computer works to a young child."]}
# ]
# response = model.generate_content(messages)


# model = genai.GenerativeModel('gemini-1.5-flash',
#                               # Set the `response_mime_type` to output JSON
#                               generation_config={"response_mime_type": "application/json"})



class CypherHelper:
    """
    see [age docs](https://age.apache.org/age-manual/master/intro/overview.html)
    """

    def __init__(self, model, db=None):
        """"""
        from funkyprompt.core import AbstractEntity

        self.model: AbstractEntity = model

    def query_from_natural_language(cls, question: str):
        """"""
        return None

    def get_graph_model_attributes(self, node):
        """the default node behviour is to just keep the name but we can 'index' other attributes
        this can either be done upfront or later on some trigger or job
        """
        return None

    # def upsert_path_query(self, node):https://age.apache.org/age-manual/master/clauses/create.html

    def create_script(self):
        """create the node - may well be a no-op but we register anyway"""
        label = self.model.get_model_fullname().replace(".", "_")
        q = f"""
        """
        return None

    def upsert_node_query(self, nodes):
        """
        create a node upsert query - any attributes can be upserted
        but labeled nodes are supposed to be unique by name
        """

        if not isinstance(nodes, list):
            nodes = [nodes]

        label = self.model.get_model_fullname().replace(".", "_")

        cypher_queries = []

        for n in nodes:
            """we may set some attributes like descriptions and stuff"""
            set_attributes = ""
            graph_attributes = self.get_graph_model_attributes(self.model)
            if graph_attributes:
                set_attributes = f"""SET n = {{
                        description: 'new desc'
                    }}"""

            cypher_queries.append(
                f"""MERGE (n:{label} {{name: '{n.name}'}})
                {set_attributes}
                RETURN n"""
            )

        """for now block them up but we investigate a more efficient notation
           we generally expect low cardinality upserts for entity types in practice
        """
        return "\n".join(cypher_queries)



from abc import ABC, abstractmethod
import typing
from . import CallingContext


class LanguageModel(ABC):

    @abstractmethod
    def get_function_call_or_stream(
        self,
        response: typing.Any,
        callback: typing.Optional[typing.Callable] = None,
        response_buffer: typing.List[typing.Any] = None,
        token_callback_action: typing.Optional[typing.Callable] = None,
    ):
        pass

    @abstractmethod
    def run(
        cls,
        messages: typing.List[dict],
        context: CallingContext,
        functions: typing.Optional[dict] = None,
    ):
        pass

    def __call__(
        cls,
        messages: typing.List[dict],
        context: CallingContext,
        functions: typing.Optional[dict] = None,
    ):
        return cls.run(context=context, messages=messages, functions=functions)



from enum import Enum


class LanguageModelProviders(Enum):

    openai = "openai"


import typing
from funkyprompt.core.agents import CallingContext
from funkyprompt.core import AbstractModel
from funkyprompt import entities


def run(
    questions: str | typing.List[str],
    context: CallingContext = None,
    model: AbstractModel = None,
):
    """entry point into the runner for convenience
    - direct questions can be asked butthen the simple `ask` method would suffice
    - the runner allows for function calling in the executor loop
    - override the default basic model with your own
    """
    from funkyprompt.core.agents import Runner

    r = Runner(model)
    return r(questions)


def ask(question, context=None, **kwargs):
    """simple model entry point
    You can ask a question just to test the basic machinery with optional streaming

    Examples
        ```python
        import funkyprompt
        from funkyprompt.core.agents import CallingContext

        a = funkyprompt.ask("what is a funkyprompt agent",
                        context = CallingContext(streaming_callback=lambda s : print(s, end='')))

        ```

        or more simply

        ```python
        funkyprompt.ask('what is the capital of ireland')
        ```

        or using different models;


    """

    from funkyprompt.services import language_model_client_from_context

    model = language_model_client_from_context()
    return model(question, context=context, **kwargs)



import typing
import pydantic


def get_pydantic_properties_string(cls):
    """
    this is useful as a prompting device
    """
    annotations = typing.get_type_hints(cls)
    class_str = f"class {cls.__name__}(BaseModel)\n"
    for field_name, field_type in annotations.items():
        field_default = getattr(cls, field_name, ...)
        field_info = cls.__fields__.get(field_name)
        description = (
            f" # {field_info.description}"
            if getattr(field_info, "description", None)
            else ""
        )
        type_str = repr(field_type)

        if field_default is ...:
            class_str += f"  -  {field_name}: {type_str}{description}\n"
        else:
            if isinstance(field_default, pydantic.Field):

                class_str += f" - {field_name}: {type_str} = Field(default={repr(field_default.default)}) {description}\n"
            else:
                class_str += f" - {field_name}: {type_str} = {repr(field_default)} {description}\n"
    return class_str



"""a map to json schema types for serialization"""

import uuid
import typing

EMBEDDING_LENGTH_OPEN_AI = 1536

PYTHON_TO_JSON_MAP = {
    str: "string",
    int: "integer",
    float: "number",
    bool: "boolean",
    list: "array",
    dict: "object",
}


def match_type(f, t):
    """a general way to see if we would match this type allowing for optionality and unions"""
    if f is t:
        return True
    args = getattr(f, "__args__", {}) or ()
    return t in args


def some_default_for_type(f):
    """it may be convenient to create a dummy value for a type
    this is somewhat subjective so we are experimenting with this over time
    this is used for DB use cases primarily so things will be serialized in such a way
    """
    if match_type(f, uuid.UUID):
        return str(uuid.uuid4())
    if match_type(f, str):
        return ""
    if match_type(f, typing.List):
        return []
    if match_type(f, dict):
        return "{}"
    if match_type(f, int):
        return 0
    if match_type(f, int):
        return 0
    if match_type(f, bool):
        return False

    return ""



"""
The postgres service is a wrapper around psycopg2 with some awareness of a pg_vector extension and age extension
In funkyprompt the game is to play nice with AbstractModels so that the postgres stuff is under the hood


# Reading
[pg_vector](https://github.com/pgvector/pgvector)
[Text Search Control](https://www.postgresql.org/docs/current/textsearch-controls.html)
"""

import json
import typing
import psycopg2
from funkyprompt.core import AbstractModel, AbstractEntity
from funkyprompt.services.data import DataServiceBase
from funkyprompt.core.utils.env import POSTGRES_CONNECTION_STRING, AGE_GRAPH
from funkyprompt.core.utils import logger
from funkyprompt.core.types.sql import VectorSearchOperator

from funkyprompt.entities import resolve as resolve_entity


def cypher_with_age_wrapper(q: str):
    """wrapper a cypher query"""
    return (
        f""" LOAD 'age';
        SET search_path = ag_catalog, "$user", public;

        SELECT * 
        FROM cypher('{AGE_GRAPH}', $$
            {q}
        $$) as (n agtype);"""
        if q
        else None
    )


def _parse_vertex_result(x):
    x = json.loads(x["n"].split("::")[0])
    model_namespace, model_name = x["label"].split("_", 1)
    name = x["properties"].get("name")
    d = {
        "entity_model_name": model_name,
        "entity_model_namespace": model_namespace,
        "name": name,
    }

    d["model"] = resolve_entity(**d)

    return d


class PostgresService(DataServiceBase):
    """the postgres service wrapper for sinking and querying entities/models

    Examples:

    ```python
    from funkyprompt.entities import Project
    #having create a model Project._register() which uses the service under the hood
    from funkyprompt.services import entity_store

    #create a typed instance
    store =entity_store(Project)

    #add entries
    p=Project(name='test', description='this is a test project for testing search about sirsh interests', labels=['test'])
    store.update_records(p)

    # look up nodes without type
    x= store.get_nodes_by_name('test')
    # specifically lookup a node of type
    store.select_one('test')

    # run any query (store.execute) or do vector search
    store.ask(question="are there any projects about sirsh's interests")

    ```
    """

    def __init__(self, model: AbstractModel):
        self.conn = psycopg2.connect(POSTGRES_CONNECTION_STRING)
        self.model: AbstractModel = model

    def _alter_model(cls):
        """try to alter the table by adding new columns only"""
        raise NotImplementedError("alter table not yet implemented")

    def _create_model(cls):
        """internal create model"""

        try:
            script = cls.model.sql().create_script()
            logger.debug(script)
            cls.execute(script)
            """
            for now create the node type separately but we could merge
            """
            script = cypher_with_age_wrapper(cls.model.cypher().create_script())
            cls.execute(script)
            logger.info(f"updated {cls.model.get_model_fullname()}")
        except Exception as pex:
            if pex is psycopg2.errors.DuplicateTable:
                cls._alter_model()
            else:
                raise

    def __drop_table__(cls):
        """drop the table - really just for testing and not something we would likely do often"""
        script = f"drop table {cls.model.get_model_fullname()}"
        logger.debug(script)
        cls.execute(script)
        logger.info(f"dropped {cls.model.get_model_fullname()}")

    def execute(
        cls,
        query: str,
        data: tuple = None,
        as_upsert: bool = False,
        page_size: int = 100,
    ):
        """run any sql query
        this works only for selects and transactional updates without selects
        """
        if not query:
            return
        try:
            c = cls.conn.cursor()
            if as_upsert:
                psycopg2.extras.execute_values(
                    c, query, data, template=None, page_size=page_size
                )
            else:
                c.execute(query, data)

            if c.description:
                result = c.fetchall()
                """if we have and updated and read we can commit and send,
                otherwise we commit outside this block"""
                cls.conn.commit()
                column_names = [desc[0] for desc in c.description or []]
                result = [dict(zip(column_names, r)) for r in result]
                return result
            """case of upsert no-query transactions"""
            cls.conn.commit()
        except Exception as pex:
            cls.conn.rollback()
            raise
        finally:
            cls.conn.close

    def execute_upsert(cls, query: str, data: tuple = None, page_size: int = 100):
        """run an upsert sql query"""
        return cls.execute(query, data=data, page_size=page_size, as_upsert=True)

    @classmethod
    def create_model(cls, model: AbstractModel):
        """creates the model based on the type.
        system fields are added for created and updated at.
        the raw table is associated with separate embeddings table via a view
        """
        return cls(model)._create_model()

    def update_records(self, records: typing.List[AbstractModel]):
        """records are updated using typed object relational mapping.
        the embedding update is queued
        """

        # TODO: there is a very confusing behaviour when the update records works on dictionaries and not objects

        if records and not isinstance(records, list):
            records = [records]
        helper = self.model.sql()
        data = [
            tuple(helper.serialize_for_db(r).values()) for i, r in enumerate(records)
        ]

        if records:
            query = helper.upsert_query(batch_size=len(records))
            result = self.execute_upsert(query=query, data=data)

            """for now do inline but this could be an async thing to not block"""
            self.queue_update_embeddings(result)

            """add the node for certain types that have unique names and are entity like
               we could in future do this as a transaction in the upserts or as a trigger 
            """
            if issubclass(self.model, AbstractEntity):
                query = cypher_with_age_wrapper(
                    self.model.cypher().upsert_node_query(records)
                )
                logger.trace(query)
                _ = self.execute(query)
                # self.queue_add_nodes(records) # or do we find a way to insert them in the insert block which would be nice
                # it seems like just adding the node as a reference with all the data is the way to do since the use case for entity lookup is one item
                # and therefore we want a fast insert and on demand we can do a two-pass resolve entities and query them
                # we could also have a slow background process that indexes the attributes we care about on the node

            return result

    def queue_update_embeddings(self, result: typing.List[dict]):
        """embeddings in general should be processed async
        when we insert some data, we read back a result with ids and column data for embeddings
        we then use whatever provided to get an embedding tensor and save it to the database
        this insert could be inline or adjacent table
        """
        from funkyprompt.core.utils.embeddings import embed_frame

        helper = self.model.sql()

        embeddings = embed_frame(
            result,
            field_mapping=self.model.get_embedding_fields(),
            id_column=helper.id_field,
        )

        query = helper.embedding_fields_partial_update_query(batch_size=len(result))

        return self.execute_upsert(
            query=query, data=(helper.partial_model_tuple(e) for e in embeddings)
        )

    def select_one(self, name: str, column: str = "name"):
        """selects one by name using the internal model"""
        table_name = self.model.get_model_fullname()
        fields = ",".join(self.model.sql().field_names)
        q = f"""SELECT { fields } FROM {table_name} where {column} = '{name}' limit 1"""
        data = self.execute(q)
        if len(data):
            return self.model(**dict(data[0]))

    def ask(self, question: str):
        """[DEPRECATE] natural language to SQL is used to query the store"""
        # prompt  = self.model.get_model_as_prompt()
        query = self.model.sql().query_from_natural_language(
            question,
        )
        return self.execute(query)

    def __getitem__(self, name: str):
        """the key value lookup on the graph is used for the labelled model type and name"""

        entity = self.select_one(name)
        """what we will do here is create what is called a wrapped entity"""
        return entity

    @classmethod
    def get_nodes_by_name(cls, name) -> typing.List[AbstractEntity]:
        """the node mode is only useful when we are invariant to types,
        because we can resolve nodes even when we dont know their type.
        Suppose an LLM knows that something _is_ an entity but does not know what it is
        we can match ANY nodes in the graph and then when we know the label->entity map
        we can then select one by one.
        We can make this more efficient in a number of ways but for now it provides a nice
        entity resolution route for agent
        Examples of optimization:
        - cached types in the instance
        - async parallel search over nodes
        """

        cypher_query = f"""MATCH (v {{name:'{name}'}}) RETURN v"""
        data = cls(AbstractEntity).query_graph(cypher_query)
        """do the entity wrapper stuff here
           should return an expanded abstract model i.e. one with lots of metadata in a structure e.g. desc, data, available functions
        """

        """we can wrap the entities when we know their type when they match name
           we need to do multiple select_ones for each matched type here (TODO:)
        """
        data = [_parse_vertex_result(x) for x in data]
        """a not so efficient way to load entities but fine for now"""
        data = [cls(d["model"]).select_one(d["name"]) for d in data]

        return data

    def query_graph(self, query: str):
        """query the graph with a valid cypher query"""
        ###
        """AGE/postgres runs cypher with some boilerplate"""
        query = cypher_with_age_wrapper(query)
        return self.execute(query)

    def ask_graph(self, question: str):
        """map the question into a valid cypher query and execute query on the graph"""
        query = self.model.cypher().query_from_natural_language(question)

        return self.query_graph(query)

    def ask(
        self,
        question: str,
        after_date: typing.Optional[dict] | str = None,
        limit: int = None,
        **kwargs,
    ):
        """
        a high level interface that determines the correct mode of query from natural language question
        different possible avenues from here... experimental
        """

        from funkyprompt.core.agents import QueryClassifier

        # TODO: this method is a pivotal point of funkyprompt - if types are good, this thing should work and be fast

        """this is experimental - there are other ways to do this e.g. push down in postgres"""
        classification: QueryClassifier = (
            QueryClassifier._classify_question_as_query_for_model(
                model=self.model, question=question
            )
        )

        print(classification)

        """we should audit all query decisions - maybe as an async task
        below this is framed as an either or to put pressure on the decision maker
        but we could also do this as a parallel search, try to avoid false fetches and combine
        """

        results = {}
        # we should try to avoid using general entity terms and use this just for specific things
        if classification.recommend_query_type == "ENTITY":
            for e in classification.entities:
                for r in self.get_nodes_by_name(e):
                    results[r.id] = r
            # if we have nothing, try the other options
            if len(results):
                return list(results.values())

        # if we have a high confidence query and it works, do this
        if (
            classification.sql_query
            and classification.sql_query_confidence_based_on_model
            > QueryClassifier.Config.MIN_QUERY_CONFIDENCE
        ):
            # TODO: manage thresholds and multiple queries
            # TODO drop embeddings from the result that is returned
            data = self.execute(classification.sql_query)
            if len(data):
                return data

        """fall back to a vector search - a temporal predicate will be needed here"""
        for q in classification.decomposed_questions:
            for r in self.vector_search(q):
                results[r["id"]] = r
            return list(results.values())

        """it may be when we try multi async routes we can dedupe and rerank here"""

        return results

    def vector_search(
        self,
        question: str,
        search_operator: VectorSearchOperator = VectorSearchOperator.INNER_PRODUCT,
        limit: int = 7,
    ):
        """
        search the model' embedding content
        in generally we can query multiple embeddings per table but for now testing with just one

        Args:
            question: a natural language question
            search_operator: the pg_vector operator type as an enum - uses the default inner product because we use the open ai embeddings by default
            limit: limit results to return

        Example:

            ```python
            from funkyprompt.core import ConversationModel
            from funkyprompt.services import entity_store
            #load the store for this type/model/table
            store =entity_store(ConversationModel)
            #search...
            store.vector_search('what did i asked about installing postgres on the mac')
            ```
        """

        from funkyprompt.core.utils.embeddings import embed_collection

        helper = self.model.sql()

        if not helper.embedding_fields:
            raise Exception(
                "this type does not support vector search as there are no embedding columns"
            )

        """default to one for now and OR later 
        - we actually need to determine the embedding provided for each column from the metadata 
        :TODO: test the more general case of multiple columns with multiple providers when getting embeddings
        it may be a different operator is better in each case
        """
        vec = embed_collection([question])[0]

        embedding_fields = helper.embedding_fields[0]
        select_fields = ",".join(helper.field_names)

        distance_max: float = (
            -0.79
        )  ##TODO: this only makes sense for the neg inner product
        part_predicates = (
            f"{embedding_fields} {search_operator.value} '{vec}' < {distance_max}"
        )

        """distances are determined in different ways, that includes what 'large' is"""
        distances = f"{embedding_fields} {search_operator.value} '{vec}' "

        """TODO: we could make some attempt to normalize for different systems
        the scale of divergence e.g. for NE_INNER_PRODUCT (-1 - d)"""
        """generate the query for now for only one embedding col"""
        query = f"""SELECT
            {select_fields},
            ({distances}) as distances
            from {helper.table_name} 
              WHERE {part_predicates}
                order by {distances} ASC LIMIT {limit}
             """

        return self.execute(query)


"""notes


"""



from funkyprompt.core import AbstractModel
from abc import ABC, abstractmethod
import typing


class DataServiceBase(ABC):
    """
    the abstract interface for all data services used in funkyprompt
    """

    @abstractmethod
    def create_model(self, model: AbstractModel):
        """
        stores can create a model which depending on the implementation is like creating a table.
        the abstract model provides all metadata for creating types in the underlying store
        """
        pass

    @abstractmethod
    def update_records(self, records: typing.List[AbstractModel], **kwargs):
        """
        data can be ingested into stores using the typed schema
        data can be queried by agents
        """
        pass

    @abstractmethod
    def select_one(self, id: str) -> AbstractModel:
        """
        a convenience to retrieve a specific entity by key
        """
        pass

    @abstractmethod
    def ask(self, question: str, **kwargs) -> typing.List[dict]:
        """
        an english question is converted into a query for the store
        each store will implement this in its own way
        an internal LLM can be used to reason about the question and convert it into the correct format eg. key-value or SQL
        """
        pass

    @abstractmethod
    def ask(self, query: str, **kwargs):
        """
        the query in the format required for the store.
        this could be a text search in a vector store, and sql query, list of keys etc.
        """



import json
import uuid
import hashlib


def uuid_str_from_dict(d):
    """
    generate a uuid string from a seed that is a sorted dict
    """
    d = json.dumps(d, sort_keys=True).encode("utf-8")
    m = hashlib.md5()
    m.update(d)
    return str(uuid.UUID(m.hexdigest()))


def funky_id(name: str = None):
    return uuid_str_from_dict({"name": name})


def funky_hash(s: str = None, length: int = 5, prefix="fun"):
    """
    generate a hash by some convention
    """
    if s:
        try:
            s = s.encode()
        except:
            pass
    s = s or str(uuid.uuid1()).encode()

    h = hashlib.shake_256(s).hexdigest(length).upper()

    return f"{prefix}{h}"



from funkyprompt.core import AbstractModel, AbstractEntity
import typing
from pydantic import Field, model_validator
import json


def create_lookup(
    data: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    """in a DAG where we are being efficient with named refs, we can get a lookup of all nodes with this and then expand for pydantic"""
    lookup = {}

    def _traverse(node: typing.Dict[str, typing.Any]):
        if "plan_description" in node:
            if node["name"] not in lookup:
                lookup[node["name"]] = {}
            lookup[node["name"]].update(node)
        if "depends" in node:
            for dep in node["depends"]:
                _traverse(dep)

    _traverse(data)
    return lookup


def expand_refs(
    node: typing.Dict[str, typing.Any],
    lookup: typing.Dict[str, typing.Dict[str, typing.Any]],
) -> typing.Dict[str, typing.Any]:
    """in a DAG where we are being efficient with named refs, we can get a lookup of all nodes with this and then expand for pydantic using this function"""
    if "depends" in node:
        expanded_depends = []
        for dep in node["depends"]:
            dep_name = dep["name"]
            if dep_name in lookup:
                expanded_dep = expand_refs(lookup[dep_name], lookup)
                expanded_depends.append(expanded_dep)
            else:
                expanded_depends.append(dep)
        node["depends"] = expanded_depends
    return node


PLANNING_PROMPT = f"""Below is a schema for building a plan.
        I would like you to consider the functions provided and build a plan to solve the task.
        You should break the problem down into steps and consider;
        (a) what functions can be used in each step matching the correct entity to the most suitable function 
        (b) what strategy to use in each step
        (c) how results from one step can be embedded as questions to the next.
        
        Functions that can be called without other dependencies can have empty dependency list and only functions that need the result of another function should depend on the other function.
        When calling functions in later states, you should always pass the shared context or known entities as parameters to the stage.
            - For example if entities are needed in other stages, they should be passed in as parameters in later stages.
        When passing data from one step to the next we can embed data in the question. 
            - For example if we got a collection of data from one or more steps, we could ask a question such as given the data [insert data] [ask the question]
            
        The user will supply further instructions for the task."""


class PlanFunctions(AbstractModel):
    name: str = Field(
        description="fully qualified function name e.g. <namespace>.<name>"
    )
    description: str = Field(
        description="a description of the function preferably with the current context taken into account e.g. provide good example parameters"
    )
    rating: float = Field(
        description="a rating from 0 to 100 for how useful this function should be in context"
    )


class Plan(AbstractEntity):
    """
    this is a base class for a plan
    a plan is something that has a question and a schema
    the plan can be chained into a dependency model
    plans in a graph should have unique names
    """

    class Config:
        name: str = "plan"
        namespace: str = "core"
        description = PLANNING_PROMPT

    name: typing.Optional[str] = Field(
        description="The unique name of the plan node", default=None
    )

    plan_description: str = Field(
        description="The plan to prompt the agent - should provide fully strategy and explain what dependencies exist with other stages"
    )
    questions: typing.Optional[typing.List[str]] = Field(
        description="The question in this plan instance as the user would ask it. A plan can be constructed without a clear question",
        default=None,
    )
    extra_arguments: typing.Optional[dict] = Field(
        description="Placeholder/hint for extra parameters that should be passed from previous stages such as data or identifiers that were discovered in the data and expected by the function either as a parameter or important context",
        default=None,
    )
    functions: typing.Optional[typing.List[PlanFunctions]] = Field(
        description="A collection of functions designed for use with this context",
        default=None,
    )
    depends: typing.Optional[typing.List["Plan"]] = Field(
        description="A dependency graph - plans can be chained into waves of functions that can be called in parallel or one after the other. Data dependencies are injected to downstream plans",
        default=None,
    )

    @model_validator(mode="before")
    @classmethod
    def _expand(cls, values):
        """expand entity refs for pydantic model"""
        l = create_lookup(values)
        values = expand_refs(values, l)
        return values

    @classmethod
    def _get_prompting_data(cls):
        """
        This is provided for promptables i.e. model schema to prompt dump.
        These data will be injected into a large prompt

        """

        from funkyprompt.core import load_entities

        return f"""
        ## Available entity functions
        
        ```json
        {
            json.dumps(load_entities(),default=str)
        }
        ```
    """

    def search_functions(cls, questions: str | typing.List[str]) -> typing.List[dict]:
        """A planner can search for new functions to answer user questions if available functions do not suffice.
           If the questions seem orthogonal they should be split into multiple questions for multiple searches.

        Args:
            questions (str | typing.List[str]): one or more questions - more is better

        Returns:
            typing.List[dict]: a list of functions and their descriptions
        """
        from funkyprompt.services import entity_store
        from funkyprompt.core.functions import Function

        return entity_store(Function).ask(questions)
